@@ -4,7 +4,7 @@
  *
  * Portions of the file below are covered by the following license:
  *
- * Copyright (c) 2003-2007 Tim Kientzle
+ * Copyright (c) 2003-2008 Tim Kientzle
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -29,7 +29,7 @@
  */
 
 #include "bsdtar_platform.h"
-__FBSDID("$FreeBSD: src/usr.bin/tar/bsdtar.c,v 1.90 2008/05/19 18:38:01 cperciva Exp $");
+__FBSDID("$FreeBSD: src/usr.bin/tar/bsdtar.c,v 1.93 2008/11/08 04:43:24 kientzle Exp $");
 
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
@@ -43,18 +43,6 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/bsdtar.c,v 1.90 2008/05/19 18:38:01 cperciva
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
-#ifdef HAVE_GETOPT_LONG
-#include <getopt.h>
-#else
-struct option {
-	const char *name;
-	int has_arg;
-	int *flag;
-	int val;
-};
-#define	no_argument 0
-#define	required_argument 1
-#endif
 #ifdef HAVE_LANGINFO_H
 #include <langinfo.h>
 #endif
@@ -88,14 +76,6 @@ struct option {
 #include "sysendian.h"
 #include "tarsnap_opt.h"
 
-#if !HAVE_DECL_OPTARG
-extern int optarg;
-#endif
-
-#if !HAVE_DECL_OPTIND
-extern int optind;
-#endif
-
 /* Global tarsnap options declared in tarsnap_opt.h. */
 int tarsnap_opt_aggressive_networking = 0;
 int tarsnap_opt_humanize_numbers = 0;
@@ -105,8 +85,6 @@ uint64_t tarsnap_opt_maxbytesout = (uint64_t)(-1);
 /* External function to parse a date/time string (from getdate.y) */
 time_t get_date(const char *);
 
-static int		 bsdtar_getopt(struct bsdtar *, const char *optstring,
-    const struct option **poption);
 static void		 build_dir(struct bsdtar *, const char *dir,
 			     const char *diropt);
 static void		 configfile(struct bsdtar *, const char *fname);
@@ -119,131 +97,6 @@ static void		 only_mode(struct bsdtar *, const char *opt,
 static void		 set_mode(struct bsdtar *, int opt, const char *optstr);
 static void		 version(void);
 
-/*
- * The leading '+' here forces the GNU version of getopt() (as well as
- * both the GNU and BSD versions of getopt_long) to stop at the first
- * non-option.  Otherwise, GNU getopt() permutes the arguments and
- * screws up -C processing.
- */
-static const char *tar_opts = "+@:BC:cdf:HhI:kLlmnOoPprtT:UvW:wX:x";
-
-/*
- * Most of these long options are deliberately not documented.  They
- * are provided only to make life easier for people who also use GNU tar.
- * The only long options documented in the manual page are the ones
- * with no corresponding short option, such as --exclude, --nodump,
- * and --fast-read.
- *
- * On systems that lack getopt_long, long options can be specified
- * using -W longopt and -W longopt=value, e.g. "-W nodump" is the same
- * as "--nodump" and "-W exclude=pattern" is the same as "--exclude
- * pattern".  This does not rely the GNU getopt() "W;" extension, so
- * should work correctly on any system with a POSIX-compliant getopt().
- */
-
-/* Fake short equivalents for long options that otherwise lack them. */
-enum {
-	OPTION_AGGRESSIVE_NETWORKING=1,
-	OPTION_CACHEDIR,
-	OPTION_CHECK_LINKS,
-	OPTION_CHROOT,
-	OPTION_EXCLUDE,
-	OPTION_FSCK,
-	OPTION_HELP,
-	OPTION_INCLUDE,
-	OPTION_HUMANIZE_NUMBERS,
-	OPTION_KEYFILE,
-	OPTION_KEEP_NEWER_FILES,
-	OPTION_LIST_ARCHIVES,
-	OPTION_LOWMEM,
-	OPTION_MAXBW,
-	OPTION_NEWER_CTIME,
-	OPTION_NEWER_CTIME_THAN,
-	OPTION_NEWER_MTIME,
-	OPTION_NEWER_MTIME_THAN,
-	OPTION_NODUMP,
-	OPTION_NO_SAME_OWNER,
-	OPTION_NO_SAME_PERMISSIONS,
-	OPTION_NOISY_WARNINGS,
-	OPTION_NUKE,
-	OPTION_NULL,
-	OPTION_ONE_FILE_SYSTEM,
-	OPTION_PRINT_STATS,
-	OPTION_SNAPTIME,
-	OPTION_STORE_ATIME,
-	OPTION_STRIP_COMPONENTS,
-	OPTION_TOTALS,
-	OPTION_VERSION,
-	OPTION_VERYLOWMEM
-};
-
-/*
- * If you add anything, be very careful to keep this list properly
- * sorted, as the -W logic relies on it.
- */
-static const struct option tar_longopts[] = {
-	{ "absolute-paths",     no_argument,       NULL, 'P' },
-	{ "aggressive-networking", no_argument,	   NULL, OPTION_AGGRESSIVE_NETWORKING },
-	{ "cachedir",		required_argument, NULL, OPTION_CACHEDIR },
-	{ "cd",                 required_argument, NULL, 'C' },
-	{ "check-links",        no_argument,       NULL, OPTION_CHECK_LINKS },
-	{ "chroot",             no_argument,       NULL, OPTION_CHROOT },
-	{ "confirmation",       no_argument,       NULL, 'w' },
-	{ "create",             no_argument,       NULL, 'c' },
-	{ "dereference",	no_argument,	   NULL, 'L' },
-	{ "directory",          required_argument, NULL, 'C' },
-	{ "exclude",            required_argument, NULL, OPTION_EXCLUDE },
-	{ "exclude-from",       required_argument, NULL, 'X' },
-	{ "extract",            no_argument,       NULL, 'x' },
-	{ "fast-read",          no_argument,       NULL, 'q' },
-	{ "file",               required_argument, NULL, 'f' },
-	{ "files-from",         required_argument, NULL, 'T' },
-	{ "fsck",		no_argument,	   NULL, OPTION_FSCK },
-	{ "help",               no_argument,       NULL, OPTION_HELP },
-	{ "humanize-numbers",	no_argument,	   NULL, OPTION_HUMANIZE_NUMBERS },
-	{ "include",            required_argument, NULL, OPTION_INCLUDE },
-	{ "interactive",        no_argument,       NULL, 'w' },
-	{ "insecure",           no_argument,       NULL, 'P' },
-	{ "keep-newer-files",   no_argument,       NULL, OPTION_KEEP_NEWER_FILES },
-	{ "keep-old-files",     no_argument,       NULL, 'k' },
-	{ "keyfile",		required_argument, NULL, OPTION_KEYFILE },
-	{ "list",               no_argument,       NULL, 't' },
-	{ "list-archives",	no_argument,	   NULL, OPTION_LIST_ARCHIVES },
-	{ "lowmem",		no_argument,	   NULL, OPTION_LOWMEM },
-	{ "maxbw",		required_argument, NULL, OPTION_MAXBW },
-	{ "modification-time",  no_argument,       NULL, 'm' },
-	{ "newer",		required_argument, NULL, OPTION_NEWER_CTIME },
-	{ "newer-ctime",	required_argument, NULL, OPTION_NEWER_CTIME },
-	{ "newer-ctime-than",	required_argument, NULL, OPTION_NEWER_CTIME_THAN },
-	{ "newer-mtime",	required_argument, NULL, OPTION_NEWER_MTIME },
-	{ "newer-mtime-than",	required_argument, NULL, OPTION_NEWER_MTIME_THAN },
-	{ "newer-than",		required_argument, NULL, OPTION_NEWER_CTIME_THAN },
-	{ "nodump",             no_argument,       NULL, OPTION_NODUMP },
-	{ "noisy-warnings",	no_argument,	   NULL, OPTION_NOISY_WARNINGS },
-	{ "norecurse",          no_argument,       NULL, 'n' },
-	{ "no-recursion",       no_argument,       NULL, 'n' },
-	{ "no-same-owner",	no_argument,	   NULL, OPTION_NO_SAME_OWNER },
-	{ "no-same-permissions",no_argument,	   NULL, OPTION_NO_SAME_PERMISSIONS },
-	{ "nuke",		no_argument,	   NULL, OPTION_NUKE },
-	{ "null",		no_argument,	   NULL, OPTION_NULL },
-	{ "one-file-system",	no_argument,	   NULL, OPTION_ONE_FILE_SYSTEM },
-	{ "preserve-permissions", no_argument,     NULL, 'p' },
-	{ "print-stats",	no_argument,	   NULL, OPTION_PRINT_STATS },
-	{ "read-full-blocks",	no_argument,	   NULL, 'B' },
-	{ "same-permissions",   no_argument,       NULL, 'p' },
-	{ "snaptime",		required_argument, NULL, OPTION_SNAPTIME },
-	{ "store-atime",	no_argument,	   NULL, OPTION_STORE_ATIME },
-	{ "strip-components",	required_argument, NULL, OPTION_STRIP_COMPONENTS },
-	{ "to-stdout",          no_argument,       NULL, 'O' },
-	{ "totals",		no_argument,       NULL, OPTION_TOTALS },
-	{ "unlink",		no_argument,       NULL, 'U' },
-	{ "unlink-first",	no_argument,       NULL, 'U' },
-	{ "verbose",            no_argument,       NULL, 'v' },
-	{ "version",            no_argument,       NULL, OPTION_VERSION },
-	{ "verylowmem",		no_argument,	   NULL, OPTION_VERYLOWMEM },
-	{ NULL, 0, NULL, 0 }
-};
-
 /* A basic set of security flags to request from libarchive. */
 #define	SECURITY					\
 	(ARCHIVE_EXTRACT_SECURE_SYMLINKS		\
@@ -253,7 +106,6 @@ int
 main(int argc, char **argv)
 {
 	struct bsdtar		*bsdtar, bsdtar_storage;
-	const struct option	*option;
 	int			 opt;
 	char			 possible_help_request;
 	char			 buff[16];
@@ -261,6 +113,7 @@ main(int argc, char **argv)
 	char			*homedir;
 	char			*conffile;
 	const char		*missingkey;
+	char			*eptr;
 
 	/*
 	 * Use a pointer for consistency, but stack-allocated storage
@@ -342,14 +195,13 @@ main(int argc, char **argv)
 	bsdtar->argv = argv;
 	bsdtar->argc = argc;
 
-	/* Process all remaining arguments now. */
 	/*
 	 * Comments following each option indicate where that option
 	 * originated:  SUSv2, POSIX, GNU tar, star, etc.  If there's
 	 * no such comment, then I don't know of anyone else who
 	 * implements that option.
 	 */
-	while ((opt = bsdtar_getopt(bsdtar, tar_opts, &option)) != -1) {
+	while ((opt = bsdtar_getopt(bsdtar)) != -1) {
 		switch (opt) {
 		case OPTION_AGGRESSIVE_NETWORKING: /* tarsnap */
 			tarsnap_opt_aggressive_networking = 1;
@@ -358,13 +210,13 @@ main(int argc, char **argv)
 			/* libarchive doesn't need this; just ignore it. */
 			break;
 		case 'C': /* GNU tar */
-			set_chdir(bsdtar, optarg);
+			set_chdir(bsdtar, bsdtar->optarg);
 			break;
 		case 'c': /* SUSv2 */
 			set_mode(bsdtar, opt, "-c");
 			break;
 		case OPTION_CACHEDIR: /* multitar */
-			bsdtar->cachedir = optarg;
+			bsdtar->cachedir = bsdtar->optarg;
 			break;
 		case OPTION_CHECK_LINKS: /* GNU tar */
 			bsdtar->option_warn_links = 1;
@@ -375,13 +227,16 @@ main(int argc, char **argv)
 		case 'd': /* multitar */
 			set_mode(bsdtar, opt, "-d");
 			break;
+		case OPTION_DRYRUN: /* tarsnap */
+			bsdtar->option_dryrun = 1;
+			break;
 		case OPTION_EXCLUDE: /* GNU tar */
-			if (exclude(bsdtar, optarg))
+			if (exclude(bsdtar, bsdtar->optarg))
 				bsdtar_errc(bsdtar, 1, 0,
-				    "Couldn't exclude %s\n", optarg);
+				    "Couldn't exclude %s\n", bsdtar->optarg);
 			break;
 		case 'f': /* multitar */
-			bsdtar->tapename = optarg;
+			bsdtar->tapename = bsdtar->optarg;
 			break;
 		case OPTION_FSCK: /* multitar */
 			set_mode(bsdtar, opt, "--fsck");
@@ -412,7 +267,7 @@ main(int argc, char **argv)
 			 * permissions without having to create those
 			 * permissions on disk.
 			 */
-			bsdtar->names_from_file = optarg;
+			bsdtar->names_from_file = bsdtar->optarg;
 			break;
 		case OPTION_INCLUDE:
 			/*
@@ -420,10 +275,10 @@ main(int argc, char **argv)
 			 * noone else needs this to filter entries
 			 * when transforming archives.
 			 */
-			if (include(bsdtar, optarg))
+			if (include(bsdtar, bsdtar->optarg))
 				bsdtar_errc(bsdtar, 1, 0,
 				    "Failed to add %s to inclusion list",
-				    optarg);
+				    bsdtar->optarg);
 			break;
 		case 'k': /* GNU tar */
 			bsdtar->extract_flags |= ARCHIVE_EXTRACT_NO_OVERWRITE;
@@ -432,7 +287,7 @@ main(int argc, char **argv)
 			bsdtar->extract_flags |= ARCHIVE_EXTRACT_NO_OVERWRITE_NEWER;
 			break;
 		case OPTION_KEYFILE: /* tarsnap */
-			load_keys(bsdtar, optarg);
+			load_keys(bsdtar, bsdtar->optarg);
 			bsdtar->have_keys = 1;
 			break;
 		case 'L': /* BSD convention */
@@ -452,10 +307,41 @@ main(int argc, char **argv)
 			bsdtar->extract_flags &= ~ARCHIVE_EXTRACT_TIME;
 			break;
 		case OPTION_MAXBW: /* tarsnap */
-			if (humansize_parse(optarg, &tarsnap_opt_maxbytesout))
+			if (humansize_parse(bsdtar->optarg,
+			    &tarsnap_opt_maxbytesout))
 				bsdtar_errc(bsdtar, 1, 0,
 				    "Cannot parse bandwidth limit: %s",
-				    optarg);
+				    bsdtar->optarg);
+			break;
+		case OPTION_MAXBW_RATE: /* tarsnap */
+			bsdtar->bwlimit_rate_up = bsdtar->bwlimit_rate_down =
+			    strtod(bsdtar->optarg, &eptr);
+			if ((*eptr != '\0') ||
+			    (bsdtar->bwlimit_rate_up < 8000) ||
+			    (bsdtar->bwlimit_rate_up > 1000000000.))
+				bsdtar_errc(bsdtar, 1, 0,
+				    "Invalid bandwidth rate limit: %s",
+				    bsdtar->optarg);
+			break;
+		case OPTION_MAXBW_RATE_DOWN: /* tarsnap */
+			bsdtar->bwlimit_rate_down =
+			    strtod(bsdtar->optarg, &eptr);
+			if ((*eptr != '\0') ||
+			    (bsdtar->bwlimit_rate_down < 8000) ||
+			    (bsdtar->bwlimit_rate_down > 1000000000.))
+				bsdtar_errc(bsdtar, 1, 0,
+				    "Invalid bandwidth rate limit: %s",
+				    bsdtar->optarg);
+			break;
+		case OPTION_MAXBW_RATE_UP: /* tarsnap */
+			bsdtar->bwlimit_rate_up =
+			    strtod(bsdtar->optarg, &eptr);
+			if ((*eptr != '\0') ||
+			    (bsdtar->bwlimit_rate_up < 8000) ||
+			    (bsdtar->bwlimit_rate_up > 1000000000.))
+				bsdtar_errc(bsdtar, 1, 0,
+				    "Invalid bandwidth rate limit: %s",
+				    bsdtar->optarg);
 			break;
 		case 'n': /* GNU tar */
 			bsdtar->option_no_subdirs = 1;
@@ -468,28 +354,28 @@ main(int argc, char **argv)
 		 * TODO: Add corresponding "older" options to reverse these.
 		 */
 		case OPTION_NEWER_CTIME: /* GNU tar */
-			bsdtar->newer_ctime_sec = get_date(optarg);
+			bsdtar->newer_ctime_sec = get_date(bsdtar->optarg);
 			break;
 		case OPTION_NEWER_CTIME_THAN:
 			{
 				struct stat st;
-				if (stat(optarg, &st) != 0)
+				if (stat(bsdtar->optarg, &st) != 0)
 					bsdtar_errc(bsdtar, 1, 0,
-					    "Can't open file %s", optarg);
+					    "Can't open file %s", bsdtar->optarg);
 				bsdtar->newer_ctime_sec = st.st_ctime;
 				bsdtar->newer_ctime_nsec =
 				    ARCHIVE_STAT_CTIME_NANOS(&st);
 			}
 			break;
 		case OPTION_NEWER_MTIME: /* GNU tar */
-			bsdtar->newer_mtime_sec = get_date(optarg);
+			bsdtar->newer_mtime_sec = get_date(bsdtar->optarg);
 			break;
 		case OPTION_NEWER_MTIME_THAN:
 			{
 				struct stat st;
-				if (stat(optarg, &st) != 0)
+				if (stat(bsdtar->optarg, &st) != 0)
 					bsdtar_errc(bsdtar, 1, 0,
-					    "Can't open file %s", optarg);
+					    "Can't open file %s", bsdtar->optarg);
 				bsdtar->newer_mtime_sec = st.st_mtime;
 				bsdtar->newer_mtime_nsec =
 				    ARCHIVE_STAT_MTIME_NANOS(&st);
@@ -516,6 +402,9 @@ main(int argc, char **argv)
 		case OPTION_NULL: /* GNU tar */
 			bsdtar->option_null++;
 			break;
+		case OPTION_NUMERIC_OWNER: /* GNU tar */
+			bsdtar->option_numeric_owner++;
+			break;
 		case 'O': /* GNU tar */
 			bsdtar->option_stdout = 1;
 			break;
@@ -556,20 +445,33 @@ main(int argc, char **argv)
 		case OPTION_SNAPTIME: /* multitar */
 			{
 				struct stat st;
-				if (stat(optarg, &st) != 0)
+				if (stat(bsdtar->optarg, &st) != 0)
 					bsdtar_errc(bsdtar, 1, 0,
-					    "Can't open file %s", optarg);
+					    "Can't open file %s",
+					    bsdtar->optarg);
 				bsdtar->snaptime = st.st_ctime;
 			}
 			break;
 		case OPTION_STORE_ATIME: /* multitar */
 			bsdtar->option_store_atime = 1;
 			break;
+		case 'S': /* NetBSD pax-as-tar */
+			bsdtar->extract_flags |= ARCHIVE_EXTRACT_SPARSE;
+			break;
+		case 's': /* NetBSD pax-as-tar */
+#if HAVE_REGEX_H
+			add_substitution(bsdtar, bsdtar->optarg);
+#else
+			bsdtar_warnc(bsdtar, 0,
+			    "-s is not supported by this version of bsdtar");
+			usage(bsdtar);
+#endif
+			break;
 		case OPTION_STRIP_COMPONENTS: /* GNU tar 1.15 */
-			bsdtar->strip_components = atoi(optarg);
+			bsdtar->strip_components = atoi(bsdtar->optarg);
 			break;
 		case 'T': /* GNU tar */
-			bsdtar->names_from_file = optarg;
+			bsdtar->names_from_file = bsdtar->optarg;
 			break;
 		case 't': /* SUSv2 */
 			set_mode(bsdtar, opt, "-t");
@@ -594,19 +496,19 @@ main(int argc, char **argv)
 #if 0
 		/*
 		 * The -W longopt feature is handled inside of
-		 * bsdtar_getop(), so -W is not available here.
+		 * bsdtar_getopt(), so -W is not available here.
 		 */
-		case 'W': /* Obscure, but useful GNU convention. */
+		case 'W': /* Obscure GNU convention. */
 			break;
 #endif
 		case 'w': /* SUSv2 */
 			bsdtar->option_interactive = 1;
 			break;
 		case 'X': /* GNU tar */
-			if (exclude_from_file(bsdtar, optarg))
+			if (exclude_from_file(bsdtar, bsdtar->optarg))
 				bsdtar_errc(bsdtar, 1, 0,
 				    "failed to process exclusions from file %s",
-				    optarg);
+				    bsdtar->optarg);
 			break;
 		case 'x': /* SUSv2 */
 			set_mode(bsdtar, opt, "-x");
@@ -616,6 +518,28 @@ main(int argc, char **argv)
 		}
 	}
 
+	/*
+	 * Sanity-check options.
+	 */
+
+	/*
+	 * Start with basic checks which can be done prior to reading
+	 * configuration files; the configuration file reading might error
+	 * out, and if someone asks for help they should get that rather than
+	 * a configuration file related error message.
+	 */
+
+	/* If no "real" mode was specified, treat -h as --help. */
+	if ((bsdtar->mode == '\0') && possible_help_request) {
+		long_help(bsdtar);
+		exit(0);
+	}
+
+	if (bsdtar->mode == '\0')
+		bsdtar_errc(bsdtar, 1, 0,
+		    "Must specify one of -c, -d, -r, -t, -x,"
+		    " --list-archives, or --print-stats");
+
 	/* Read options from configuration files. */
 	if ((homedir = getenv("HOME")) != NULL) {
 		if (asprintf(&conffile, "%s/.tarsnaprc", homedir) == -1)
@@ -628,20 +552,7 @@ main(int argc, char **argv)
 	}
 	configfile(bsdtar, ETC_TARSNAP_CONF);
 
-	/*
-	 * Sanity-check options.
-	 */
-
-	/* If no "real" mode was specified, treat -h as --help. */
-	if ((bsdtar->mode == '\0') && possible_help_request) {
-		long_help(bsdtar);
-		exit(0);
-	}
-
-	if (bsdtar->mode == '\0')
-		bsdtar_errc(bsdtar, 1, 0,
-		    "Must specify one of -c, -d, -r, -t, -x,"
-		    " --list-archives, or --print-stats");
+	/* Continue with more sanity-checking. */
 	if ((bsdtar->tapename == NULL) &&
 	    (bsdtar->mode != OPTION_PRINT_STATS &&
 	     bsdtar->mode != OPTION_LIST_ARCHIVES &&
@@ -659,6 +570,14 @@ main(int argc, char **argv)
 	if (bsdtar->have_keys == 0)
 		bsdtar_errc(bsdtar, 1, 0,
 		    "Keys must be provided via --keyfile option");
+	if (tarsnap_opt_aggressive_networking != 0) {
+		if ((bsdtar->bwlimit_rate_up != 0) ||
+		    (bsdtar->bwlimit_rate_down != 0)) {
+			bsdtar_errc(bsdtar, 1, 0,
+			    "--aggressive-networking is incompatible with"
+			    " --maxbw-rate options");
+		}
+	}
 
 	/*
 	 * The -f option doesn't make sense for --list-archives, --fsck, or
@@ -689,6 +608,8 @@ main(int argc, char **argv)
 		only_mode(bsdtar, "--maxbw", "c");
 	if (bsdtar->option_dont_traverse_mounts)
 		only_mode(bsdtar, "--one-file-system", "c");
+	if (bsdtar->option_dryrun)
+		only_mode(bsdtar, "--dry-run", "c");
 	if (bsdtar->option_fast_read)
 		only_mode(bsdtar, "--fast-read", "xt");
 	if (bsdtar->option_honor_nodump)
@@ -720,9 +641,6 @@ main(int argc, char **argv)
 	if (bsdtar->strip_components != 0)
 		only_mode(bsdtar, "--strip-components", "xt");
 
-	bsdtar->argc -= optind;
-	bsdtar->argv += optind;
-
 	/*
 	 * Canonicalize the path to the cache directories.  This is
 	 * necessary since the tar code can change directories.
@@ -767,6 +685,13 @@ main(int argc, char **argv)
 		    "The %s key is required for %s but is not available",
 		    missingkey, bsdtar->modestr);
 
+	/* Tell the network layer how much bandwidth to use. */
+	if (bsdtar->bwlimit_rate_up == 0)
+		bsdtar->bwlimit_rate_up = 1000000000.;
+	if (bsdtar->bwlimit_rate_down == 0)
+		bsdtar->bwlimit_rate_down = 1000000000.;
+	network_bwlimit(bsdtar->bwlimit_rate_down, bsdtar->bwlimit_rate_up);
+
 	/* Perform the requested operation. */
 	switch(bsdtar->mode) {
 	case 'c':
@@ -799,6 +724,9 @@ main(int argc, char **argv)
 	}
 
 	cleanup_exclusions(bsdtar);
+#if HAVE_REGEX_H
+	cleanup_substitution(bsdtar);
+#endif
 
 	/* Clean up network layer. */
 	network_fini();
@@ -861,11 +789,7 @@ usage(struct bsdtar *bsdtar)
 	     " [filenames...]\n", p);
 	fprintf(stderr, "  Delete:     %s [options...] -df <archive>\n", p);
 	fprintf(stderr, "  Tar output: %s [options...] -rf <archive>\n", p);
-#ifdef HAVE_GETOPT_LONG
 	fprintf(stderr, "  Help:    %s --help\n", p);
-#else
-	fprintf(stderr, "  Help:    %s -h\n", p);
-#endif
 	exit(1);
 }
 
@@ -887,11 +811,7 @@ static const char *long_help_msg =
 	"  -w    Interactive\n"
 	"Create: %p -c [options] [<file> | <dir> | @<archive> | -C <dir> ]\n"
 	"  <file>, <dir>  add these items to archive\n"
-#ifdef HAVE_GETOPT_LONG
 	"  --exclude <pattern>  Skip files that match pattern\n"
-#else
-	"  -W exclude=<pattern>  Skip files that match pattern\n"
-#endif
 	"  -C <dir>  Change to <dir> before processing remaining files\n"
 	"  @<filename>  Add entries from archive <filename>\n"
 	"  @@ <archive>  Add entries from tarsnap archive <archive>\n"
@@ -931,83 +851,6 @@ long_help(struct bsdtar *bsdtar)
 	version();
 }
 
-static int
-bsdtar_getopt(struct bsdtar *bsdtar, const char *optstring,
-    const struct option **poption)
-{
-	char *p, *q;
-	const struct option *option;
-	int opt;
-	int option_index;
-	size_t option_length;
-
-	option_index = -1;
-	*poption = NULL;
-
-#ifdef HAVE_GETOPT_LONG
-	opt = getopt_long(bsdtar->argc, bsdtar->argv, optstring,
-	    tar_longopts, &option_index);
-	if (option_index > -1)
-		*poption = tar_longopts + option_index;
-#else
-	opt = getopt(bsdtar->argc, bsdtar->argv, optstring);
-#endif
-
-	/* Support long options through -W longopt=value */
-	if (opt == 'W') {
-		p = optarg;
-		q = strchr(optarg, '=');
-		if (q != NULL) {
-			option_length = (size_t)(q - p);
-			optarg = q + 1;
-		} else {
-			option_length = strlen(p);
-			optarg = NULL;
-		}
-		option = tar_longopts;
-		while (option->name != NULL &&
-		    (strlen(option->name) < option_length ||
-		    strncmp(p, option->name, option_length) != 0 )) {
-			option++;
-		}
-
-		if (option->name != NULL) {
-			*poption = option;
-			opt = option->val;
-
-			/* If the first match was exact, we're done. */
-			if (strncmp(p, option->name, strlen(option->name)) == 0) {
-				while (option->name != NULL)
-					option++;
-			} else {
-				/* Check if there's another match. */
-				option++;
-				while (option->name != NULL &&
-				    (strlen(option->name) < option_length ||
-				    strncmp(p, option->name, option_length) != 0)) {
-					option++;
-				}
-			}
-			if (option->name != NULL)
-				bsdtar_errc(bsdtar, 1, 0,
-				    "Ambiguous option %s "
-				    "(matches both %s and %s)",
-				    p, (*poption)->name, option->name);
-
-			if ((*poption)->has_arg == required_argument
-			    && optarg == NULL)
-				bsdtar_errc(bsdtar, 1, 0,
-				    "Option \"%s\" requires argument", p);
-		} else {
-			opt = '?';
-			/* TODO: Set up a fake 'struct option' for
-			 * error reporting... ? ? ? */
-		}
-	}
-
-	return (opt);
-}
-
 static void
 build_dir(struct bsdtar *bsdtar, const char *dir, const char *diropt)
 {
@@ -1081,6 +924,8 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 	char * conf_opt;
 	char * conf_arg;
 	size_t optlen;
+	char * homedir;
+	char * conf_arg_malloced;
 
 	/* Ignore comments and blank lines. */
 	if ((line[0] == '#') || (line[0] == '\0'))
@@ -1113,6 +958,28 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 		conf_arg = NULL;
 	}
 
+	/*
+	 * If we have an argument which starts with ~, and ${HOME} is
+	 * defined, expand ~ to $HOME.
+	 */
+	if ((conf_arg != NULL) && (conf_arg[0] == '~') &&
+	    ((homedir = getenv("HOME")) != NULL)) {
+		/* Allocate space for the expanded argument string. */
+		if ((conf_arg_malloced =
+		    malloc(strlen(conf_arg) + strlen(homedir))) == NULL)
+			bsdtar_errc(bsdtar, 1, errno, "Out of memory");
+
+		/* Copy $HOME and the argument sans leading ~. */
+		memcpy(conf_arg_malloced, homedir, strlen(homedir));
+		memcpy(conf_arg_malloced + strlen(homedir), &conf_arg[1],
+		    strlen(&conf_arg[1]) + 1);
+
+		/* Use the expanded argument string hereafter. */
+		conf_arg = conf_arg_malloced;
+	} else {
+		conf_arg_malloced = NULL;
+	}
+
 	if (strcmp(conf_opt, "aggressive-networking") == 0) {
 		tarsnap_opt_aggressive_networking = 1;
 	} else if (strcmp(conf_opt, "cachedir") == 0) {
@@ -1186,6 +1053,9 @@ configfile_helper(struct bsdtar *bsdtar, const char *line)
 		    conf_opt);
 	}
 
+	/* Free expanded argument or NULL. */
+	free(conf_arg_malloced);
+
 	/* Free memory allocated by strdup. */
 	free(conf_opt);
 
